/*
 * $Id: DelegationPoint.java 256 2005-08-30 19:21:22Z davidb $
 * 
 * Copyright (c) 2005 VeriSign. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The name of the author may not
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package se.rfc.unbound;

import java.net.InetAddress;
import java.util.*;

import org.apache.log4j.Logger;
import org.xbill.DNS.*;

import se.rfc.unbound.cache.HostCache;

/**
 * This is a class representing a delegation point. It is essentially the NS
 * rrset for the delegation and the corresponding A or AAAA rrsets for the NS
 * record targets. Its primary purpose is to track the use and availability of
 * nameservers for the given delegation.
 * 
 * @author davidb
 */
public class DelegationPoint
{

  /**
   * The standalone name of the delegation point. This may be set if the
   * delegation point was created without an NS rrset.
   */
  private Name          mName;

  /** This is the defining delegation RRset for this delegation point. */
  private RRset         mNSRRset;
  /**
   * This is a mapping of target names to Boolean values indicating whether or
   * not the target has been resolved. (i.e., IP addresses for that name have
   * been added).
   */
  private Map           mTargetNameMap;
  /**
   * This is a list of Host object representing the actual targets.
   */
  private List          mTargets;
  /**
   * This is a list of Host objects that have yet to be tried.
   */
  private List          mUsableTargets;

  private Logger        log              = Logger.getLogger(this.getClass());

  // This just exists to convert the Lists to arrays.
  private static Host[] empty_host_array = new Host[0];

  /**
   * The core constructor.
   */
  private DelegationPoint()
  {
    mName = null;
    mTargetNameMap = new HashMap();
    mTargets = new ArrayList();
    mUsableTargets = new ArrayList();
  }

  /**
   * Create a DelegationPoint based on a NS rrset.
   * 
   * @param ns_rrset The NS rrset.
   */
  public DelegationPoint(RRset ns_rrset)
  {
    this();

    mNSRRset = ns_rrset;

    // Load the targetNameMap
    for (Iterator i = ns_rrset.rrs(); i.hasNext();)
    {
      NSRecord ns = (NSRecord) i.next();
      mTargetNameMap.put(ns.getTarget(), Boolean.FALSE);
    }
  }

  /**
   * Create a DelegationPoint based on a name and list of Host objects. This
   * sort of DelegationPoint cannot have targets added to it later.
   * 
   * @param n The name of the delegation point.
   * @param targetList A list of Host objects.
   */
  public DelegationPoint(Name n, List targetList)
  {
    this();

    mNSRRset = null;
    mName = n;
    mTargets = targetList;
    mUsableTargets.addAll(targetList);
  }

  /** @return the name of this delegation. */
  public Name getName()
  {
    if (mName != null) return mName;
    return mNSRRset.getName();
  }

  /**
   * Add a found IP address to the delegation point. This will only succeed if
   * the name, n, is one of the known target names for this delegation.
   * 
   * @param n The name associated with the target (i.e., the ownername of the
   *          A/AAAA record)
   * @param h The host object associated with the target.
   */
  public void addTarget(Name n, Host h)
  {
    if (h == null) throw new NullPointerException();
    if (n == null) throw new NullPointerException();

    if (mTargetNameMap.size() == 0) return;

    if (!mTargetNameMap.containsKey(n))
    {
      log.warn("Not adding irrelevant target to " + getName() + ": " + n
          + ", " + h);
      return;
    }

    mTargets.add(h);
    mUsableTargets.add(h);
    mTargetNameMap.put(n, Boolean.TRUE);
  }

  /**
   * Set an otherwise missing target as "failed" -- this mainly means that it
   * will not show up as a missing name.
   * 
   * @param n The name to fail.
   */
  public void failTarget(Name n)
  {
    if (mTargetNameMap.containsKey(n))
    {
      mTargetNameMap.put(n, Boolean.TRUE);
    }
  }
  
  /**
   * Set an otherwise missing target as "active" -- this mainly means that it
   * will not show up as a missing name.
   * 
   * @param n The name to mark.
   */
  public void markTarget(Name n)
  {
    if (mTargetNameMap.containsKey(n))
    {
      mTargetNameMap.put(n, Boolean.TRUE);
    }
  }
  
  /**
   * Add a target based on the (address) record r. This will lookup or place
   * an entry in the host cache, if provided.
   * 
   * @param r An address record to use as a nameserver target.
   * @param hostCache An (optional) reference to a host cache.
   */
  private void addTarget(Record r, HostCache hostCache)
  {
    InetAddress addr;

    if (r.getType() == Type.A)
    {
      addr = ((ARecord) r).getAddress();
    }
    else if (r.getType() == Type.AAAA)
    {
      addr = ((AAAARecord) r).getAddress();
    }
    else
    {
      // Some other type? Nothing to do.
      return;
    }

    Host h = null;
    if (hostCache != null)
    {
      h = hostCache.lookup(addr);
    }
    else
    {
      h = new Host(addr);
    }

    addTarget(r.getName(), h);
  }

  /**
   * Add IP addresses to the delegation point. This will only succeed if the
   * RRset is of type A or AAAA and the record name matches a NS record
   * target.
   * 
   * @param rrset The RRset to add.
   * @param hostCache This may be null, but if it is present, the IP addresses
   *          will be looked up/placed in the host cache.
   */
  public void addTargets(RRset rrset, HostCache hostCache)
  {
    if (rrset == null) throw new NullPointerException();

    int rrset_type = rrset.getType();
    if (rrset_type != Type.A && rrset_type != Type.AAAA) return;

    for (Iterator i = rrset.rrs(); i.hasNext();)
    {
      Record r = (Record) i.next();
      addTarget(r, hostCache);
    }
  }

  /**
   * Add an array of target RRsets.
   * 
   * @param rrsets The array of address-type RRsets.
   * @param hostCache This may be null, but, if present, the host objects will
   *          be looked up/placed in the host cache.
   */
  public void addTargets(RRset[] rrsets, HostCache hostCache)
  {
    for (int i = 0; i < rrsets.length; i++)
      addTargets(rrsets[i], hostCache);
  }

  public boolean hasTargetName(Name n)
  {
    return mTargetNameMap.containsKey(n);
  }

  /**
   * @return An array of Host object representing nameserver targets. This
   *         will only return targets that it has not previously returned.
   */
  public Host[] getTargets()
  {
    Host[] res = (Host[]) mUsableTargets.toArray(empty_host_array);

    mUsableTargets.clear();

    return res;
  }

  /**
   * Reset the target list so that getTargets() will return all known targets
   * for this delegation point, regardless of whether or not the target has
   * been returned. This is mostly useful if reusing a DelegationPoint object.
   */
  public void resetTargets()
  {
    mUsableTargets.clear();
    mUsableTargets.addAll(mTargets);
  }

  /**
   * @return A list of Names of delegation point target names with no IP
   *         addresses.
   */
  public List getMissingTargetNames()
  {
    List res = new LinkedList();

    for (Iterator i = mTargetNameMap.entrySet().iterator(); i.hasNext();)
    {
      Map.Entry entry = (Map.Entry) i.next();
      boolean v = ((Boolean) entry.getValue()).booleanValue();
      if (!v) res.add(entry.getKey());
    }

    return res;
  }

  /**
   * @return true if there are delegation point hostnames without IP
   *         addresses.
   */
  public boolean hasMissingTargets()
  {
    if (mTargetNameMap.size() == 0) return false;

    for (Iterator i = mTargetNameMap.values().iterator(); i.hasNext();)
    {
      boolean v = ((Boolean) i.next()).booleanValue();
      if (!v) return true;
    }

    return false;
  }

  public String toString()
  {
    StringBuffer sb = new StringBuffer();

    sb.append("DelegationPoint <");
    sb.append(getName());
    sb.append("> [");

    if (mNSRRset != null)
    {
      for (Iterator i = mTargetNameMap.entrySet().iterator(); i.hasNext();)
      {
        Map.Entry e = (Map.Entry) i.next();
        boolean v = ((Boolean) e.getValue()).booleanValue();
        if (v) sb.append('*');
        sb.append(e.getKey());
        if (i.hasNext()) sb.append(' ');
      }
    }
    else
    {
      for (Iterator i = mTargets.iterator(); i.hasNext();)
      {
        sb.append(i.next().toString());
        if (i.hasNext()) sb.append(' ');
      }
    }
    sb.append("]");

    return sb.toString();
  }

  /**
   * Given a referral message, construct a DelegationPoint object.
   * 
   * Note that this method does not actually test to see if the message is an
   * actual referral. It really is just checking to see if it can construct a
   * delegation point, so the message could be of some other type (some ANSWER
   * messages, some CNAME messages, generally.) Note that the resulting
   * DelegationPoint will contain targets for all "relevant" glue (i.e.,
   * address records whose ownernames match the target of one of the NS
   * records), so if policy dictates that some glue should be discarded beyond
   * that, discard it before calling this method. Note that this method will
   * find "glue" in either the ADDITIONAL section or the ANSWER section.
   * 
   * @param m the (REFERRAL) message.
   * @param hostCache The host cache, if there is one. Targets will be looked
   *          up in the host cache, if one is provided.
   * 
   * @return A DelegationPoint representing this referral, or null if the
   *         message was not appropriate.
   */
  public static DelegationPoint messageToDelegationPoint(SMessage m,
      HostCache hostCache)
  {
    DelegationPoint dp = null;
    SRRset ns_rrset = null;

    // Look for NS records in the authority section...
    SRRset[] rrsets = m.getSectionRRsets(Section.AUTHORITY, Type.NS);
    if (rrsets.length > 0) ns_rrset = rrsets[0];

    // In some cases (even legitimate, perfectly legal cases), the NS set for
    // the "referral" might be in the answer section.
    if (ns_rrset == null)
    {
      rrsets = m.getSectionRRsets(Section.ANSWER, Type.NS);
      if (rrsets.length > 0) ns_rrset = rrsets[0];
    }

    // If there was no NS rrset in the authority section, then this wasn't a
    // referral message. (It might not actually be a referral message anyway)
    if (ns_rrset == null) return null;

    // If we found any, then Yay! we have a delegation point.
    dp = new DelegationPoint(ns_rrset);

    // Load glue targets. To be thorough, we will look for "glue" in the
    // ANSWER section as well.
    rrsets = m.getSectionRRsets(Section.ANSWER, Type.A);
    dp.addTargets(rrsets, hostCache);
    rrsets = m.getSectionRRsets(Section.ANSWER, Type.AAAA);
    dp.addTargets(rrsets, hostCache);

    rrsets = m.getSectionRRsets(Section.ADDITIONAL, Type.A);
    dp.addTargets(rrsets, hostCache);
    rrsets = m.getSectionRRsets(Section.ADDITIONAL, Type.AAAA);
    dp.addTargets(rrsets, hostCache);

    return dp;
  }
}