
package se.rfc.unbound.server;

import java.net.InetSocketAddress;
import java.nio.ByteBuffer;

import se.rfc.unbound.DNSEvent;

/**
 * This class represents a request to write a message over a given channel. It
 * contains all the information needed to write the actual message as well as
 * do bookeeping after the fact.
 * 
 * @author davidb
 */
public class WriteEvent
{
  /** The message to write, in its entirety */
  private ByteBuffer        mData;
  /**
   * The destination to write the message to. If this is attached to a
   * (connected) TCP channel, this will be ignored.
   */
  private InetSocketAddress mDestination;

  /**
   * An opaque object that (ostensibly) is used to match responses to this
   * request (if it is a request).
   */
  private Object            mMatchState;

  /**
   * If this is a request, then this is the event that ulitimately will be
   * reactivated when matched with a response. Basically, this is just being
   * passed through.
   */
  private DNSEvent          mEvent;

  /**
   * Generate a new WriteEvent.
   * 
   * @param data The data to write.
   * @param destination The destination to write the data to.
   * @param event The event, may be null.
   * @param m The request message, may be null.
   */
  public WriteEvent(byte[] data, InetSocketAddress destination,
      DNSEvent event, Object matchState)
  {
    mData = ByteBuffer.wrap(data);
    mDestination = destination;
    mEvent = event;
    mMatchState = matchState;
  }

  /**
   * @return The data that should be written.
   */
  public ByteBuffer getData()
  {
    return mData;
  }

  /**
   * @return Returns the destination.
   */
  public InetSocketAddress getDestination()
  {
    return mDestination;
  }

  /**
   * @return Returns the matchState.
   */
  public Object getMatchState()
  {
    return mMatchState;
  }

  public boolean isRequest()
  {
    return (mMatchState != null);
  }

  /**
   * @return Returns the event.
   */
  public DNSEvent getEvent()
  {
    return mEvent;
  }
  
  public boolean isClosed()
  {
    if (mEvent != null) return mEvent.isClosed();
    return false;
  }
}